
import java.net.*;
import java.io.*;
import java.nio.*;
import java.util.regex.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.text.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.border.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.nio.channels.*;
import java.awt.image.*;
import javax.imageio.*;
import javax.net.*;
import javax.net.ssl.*;
import java.security.*;
import java.util.jar.*;
import java.security.cert.*;
import java.math.*;

import org.jdom.*;
import org.jdom.adapters.*;
import org.jdom.filter.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.transform.*;
import org.jdom.xpath.*;

public class Board {
    
    public static void main(String[] args) throws Exception {
        for (int i = 0; i < 40; i++) {
            Board b = new Board(15, 15, 0.25);
            b.save("boards/" + b.difficulty + ".xml");
        }
        //~ Board b = new Board("haha.xml");
        //~ BoardVerifier.verify(b);
    }
    
    public int difficulty = 0;
    public int width;
    public int height;
    public Set<Spot> mines = new HashSet<Spot>();
    public Spot start;
        
    public Board(String filename) throws Exception {
        this(u.getElement(u.loadXml(filename), "//board"));
    }
        
    public Board(File file) throws Exception {
        this(u.getElement(u.loadXml(file), "//board"));
    }
    
    public Board(int width, int height, double mineChance) throws Exception {
        this.width = width;
        this.height = height;
        
        boolean found = false;
        for (int i = 0; i < 1000; i++) {
            layMines(mineChance);
            pickStart();
            if (verify()) {
                found = true;
                break;
            }
        }
        if (!found) {
            throw new IllegalArgumentException("failed to make a board");
        }
    }
    
    public Board(Element e) {
        fromElement(e);
    }
    
    public void save(String filename) throws Exception {
        u.save(new File(filename), u.makeDocument(toElement()));
    }
    
    public void fromElement(Element e) {
        try {
            difficulty = new Integer(e.getAttributeValue("difficulty"));
        } catch (Exception ee) {
        }
        fromString(((CDATA)e.getContent(0)).getText());
    }
    
    public Element toElement() {
        Element e = new Element("board");
        e.setAttribute("difficulty", "" + difficulty);
        e.addContent(new CDATA(toString()));
        return e;
    }
    
    public void fromString(String s) {
        Matcher m = Pattern.compile("(?s)start\\s*(.*?)\\s*end").matcher(s);
        if (m.find()) {
            String[] lines = m.group(1).split("\\s+");
            if (lines.length == 0 || lines[0].length() == 0) {
                throw new IllegalArgumentException("invalid board size: " + m.group(1));
            }
            
            width = lines[0].length();
            height = lines.length;
            for (int y = 0; y < height; y++) {
                String line = lines[y];
                for (int x = 0; x < width; x++) {
                    Spot spot = new Spot(x, y);
                    switch (line.charAt(x)) {
                        case 'M' : layMine(spot); break;
                        case 'S' : start = spot; break;
                    }
                }
            }
        } else {
            throw new IllegalArgumentException("invalid board representation: " + s);
        }
    }
    
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("start\n");
        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                Spot spot = new Spot(x, y);
                if (mines.contains(spot)) {
                    buf.append("M");
                } else if (spot.equals(start)) {
                    buf.append("S");
                } else {
                    buf.append("-");
                }
            }
            buf.append("\n");
        }
        buf.append("end\n");
        return buf.toString();
    }
    
    public boolean verify() throws Exception {
        difficulty = BoardVerifier.verify(this);
        return difficulty > 0;
    }
    
    public void clearMines() {
        mines.clear();
    }
    
    public void layMines(double mineChance) {
        if (mineChance >= 0.9) {
            throw new IllegalArgumentException("mine chance too high");
        }
        
        while (true) {
            clearMines();
            for (int x = 0; x < width; x++) {
                for (int y = 0; y < height; y++) {
                    if (u.r.nextDouble() < mineChance) {
                        layMine(x, y);
                    }
                }
            }
            if (mines.size() < width * height) {
                break;
            }
        }
    }
    
    public void pickStart() {
        if (width * height == mines.size()) {
            throw new IllegalArgumentException("too many mines... nowhere to start");
        }
        
        while (true) {
            start = new Spot(u.r.nextInt(width), u.r.nextInt(height));
            if (!mines.contains(start)) {
                break;
            }
        }
    }
    
    public boolean onBoard(Spot spot) {
        return
            spot.x >= 0 && spot.x < width &&
            spot.y >= 0 && spot.y < height;
    }
    
    public void layMine(int x, int y) {
        layMine(new Spot(x, y));
    }
    
    public void layMine(Spot spot) {
        mines.add(spot);
    }
    
    public boolean mine(Spot spot) {
        return mines.contains(spot);
    }
    
    public int squareCount() {
        return width * height;
    }
    
    public Vector<Spot> adjacentMines(Spot center) {
        Vector<Spot> v = new Vector<Spot>();
        for (Spot spot : adjacent(center)) {
            if (mines.contains(spot)) {
                v.add(spot);
            }
        }
        return v;
    }
    
    public int mineCount(Spot center) {
        return adjacentMines(center).size();
    }
    
    public Vector<Spot> adjacent(Spot center) {
        Vector<Spot> adjacent = new Vector<Spot>();
        for (int x = -1; x <= 1; x++) {
            for (int y = -1; y <= 1; y++) {
                if (x == 0 && y == 0) continue;
                    
                Spot spot = new Spot(center.x + x, center.y + y);
                if (onBoard(spot)) {                
                    adjacent.add(spot);
                }
            }
        }
        return adjacent;
    }
    
    public Spot toSymmetricSpace(Spot spot, int otherSpace) {
        int x = spot.x;
        int y = spot.y;
        if ((otherSpace & 1) != 0) {
            x = width - 1 - x;
        }
        if ((otherSpace & 2) != 0) {
            y = height - 1 - y;
        }
        if ((otherSpace & 4) != 0) {
            return new Spot(y, x);
        } else {
            return new Spot(x, y);
        }
    }
    
    public Spot fromSymmetricSpace(Spot spot, int otherSpace) {
        int x = spot.x;
        int y = spot.y;
        if ((otherSpace & 4) != 0) {
            x = spot.y;
            y = spot.x;
        }
        if ((otherSpace & 1) != 0) {
            x = width - 1 - x;
        }
        if ((otherSpace & 2) != 0) {
            y = height - 1 - y;
        }
        return new Spot(x, y);
    }
}
